# -*- coding: utf-8 -*-
"""Tools for celestial mechanics."""

import sympy # sympy autodetects python 3 here, but errors when importing from other .py files in package

# Find suffix
import sysconfig
suffix = sysconfig.get_config_var('EXT_SUFFIX')
if suffix is None:
    suffix = ".so"

# Import shared library
import os
import warnings
pymodulepath = os.path.dirname(__file__)
from ctypes import cdll, c_char_p
clibcelmech = cdll.LoadLibrary(pymodulepath+"/../libcelmech"+suffix)

# Version
__version__ = c_char_p.in_dll(clibcelmech, "celmech_version_str").value.decode('ascii')

# Build
__build__ = c_char_p.in_dll(clibcelmech, "celmech_build_str").value.decode('ascii')

# Githash
__githash__ = c_char_p.in_dll(clibcelmech, "celmech_githash_str").value.decode('ascii')

# Check for version
try:
    moduleversion = pkg_resources.require("celmech")[0].version
    libcelmechversion = __version__
    if moduleversion != libcelmechversion:
        warnings.warn("WARNING: python module and libcelmech have different version numbers: '%s' vs '%s'.\n" %(moduleversion, libcelmechversion), ImportWarning)
except:
    # Might fails on python3 versions, but not important
    pass

from .andoyer import Andoyer, AndoyerHamiltonian
from .poincare import PoincareParticle, Poincare, PoincareHamiltonian

__all__ = ["Andoyer", "AndoyerHamiltonian", "PoincareParticle", "Poincare", "PoincareHamiltonian"]



from sympy import S, diff, lambdify, symbols, sqrt, cos,sin, numbered_symbols, simplify,binomial, hyper, hyperexpand, Function, factorial,elliptic_k,elliptic_e, expand_trig
from . import clibcelmech
from ctypes import Structure, c_double, POINTER, c_float, c_int, c_uint, c_uint32, c_int64, c_long, c_ulong, c_ulonglong, c_void_p, c_char_p, CFUNCTYPE, byref, create_string_buffer, addressof, pointer, cast
from scipy.integrate import quad
import math
import numpy as np
from scipy.optimize import leastsq


def laplace_coefficient(s,j,n,a):
    """
    Calculates nth derivative with respect to a (alpha) of Laplace coefficient b_s^j(a).
    Code due to Jack Wisdom. Changed notation to match Murray & Dermott (Eq. 6.67)
    
    Arguments
    ---------
    s : float 
        half-integer parameter of Laplace coefficient. 
    j : int 
        integer parameter of Laplace coefficient. 
    n : int 
        return nth derivative with respect to a of b_s^j(a)
    a : float
        semimajor axis ratio a1/a2 (alpha)
    """    
    clibcelmech.laplace.restype = c_double
    return clibcelmech.laplace(c_double(s), c_int(j), c_int(n), c_double(a))

def general_order_coefficient(res_j, order, epower, a):
    clibcelmech.GeneralOrderCoefficient.restype = c_double
    return clibcelmech.GeneralOrderCoefficient(c_int(res_j), c_int(order), c_int(epower), c_double(a))

# Vector of resonance coefficients
def get_res_coeff_vector(j,k):
    """Returns a vector comprised of all sub-resonance coefficients for the j:j-k mean motion resonance""" 
    res_pratio = float(j - k) /float(j)
    alpha = res_pratio**(2./3.)
    Cjkl = general_order_coefficient
    return np.array([Cjkl(j,k,l,alpha) for l in range(k+1)],dtype=np.float64)

def get_fg_coeffs(res_j,res_k):
    """Get 'f' and 'g' coefficients for approximating the disturbing function coefficients associated with an MMR."""
    res_pratio = float(res_j - res_k) /float(res_j)
    alpha = res_pratio**(2./3.)
    vec = get_res_coeff_vector(res_j,res_k)
    resids_vec_fn = lambda fg: vec - np.array([binomial(res_k,l) * fg[0]**(l) * fg[1]**(res_k-l) for l in range(res_k+1)],dtype=np.float64)
    ex = (1-alpha)
    f0 = -1 / ex
    g0 = 1 / ex
    f,g = leastsq(resids_vec_fn,(f0,g0))[0]
    return f,g

def Xlm0(l,m,e):
    """ 
        Get the closed-form expression for the Hansen coefficient X^(l,m)_0
        which appears in the secular component of the disturbing function (e.g., Mardling 2013) 
    """
    a =  (m-l-1) / S(2)
    b = a + 1/ S(2)
    c = m + 1
    fn = hyperexpand(hyper([a,b],[c],e*e))
    return (-e / 2)**m * binomial(l+m+1,m) * fn

def secular_eps_l_Df_m(l,m,e,e1):
    """
        Get the secularly-averaged value of 
            eps^l * exp[ i*m*(f'-f) ] 
        where l and m are integers, f' and f are true anamolies, and
            eps =  (r/a) / (r'/a')  - 1
        where r and a are radial distance and semi-major axis.
    """
    s = 0
    for i in range(l+1):
        s = s + binomial(l,i)*(-1)**(l-i) * Xlm0(i,m,e) * Xlm0(-i-1,m,e1)
    return s
def secular_DF_harmonic_term(e,e1,dw,m,order):
    s = 0
    b = Function('b') 
    alpha = symbols('alpha')
    if order%2==0:
        maxk = order + 1
    else:
        maxk=order
    for i in range(maxk):
     s = s + alpha**i *  b(1/S(2),m,i,alpha) / factorial(i) * secular_eps_l_Df_m(i,m,e,e1)
    if m==0:
        s=s / S(2)
    return s * cos(m*dw)
def secular_DF(e,e1,w,w1,order):
    """ 
    Return the secular component of the disturbing function for coplanet planets up to
    the specified order in the planets' eccentricities for an inner plaent with eccentricity 
    and longitude of periapse (e,w) and outer planet eccentricity/longitude of periape (e1,w1)
    """
    s = 0
    dw = w1 - w
    # Introduce `eps' as order parameter multiplying eccentricities. 
    # 'eps' is set to 1 at teh end of calculation.
    eps = S('epsilon')
    maxM = int( np.floor( order / 2.) )
    for i in range(maxM+1):
        term = secular_DF_harmonic_term(eps*e,eps*e1,dw,i,order) 
        term = term.series(eps,0,order+1).removeO()
        s = s + expand_trig(term) 
    return s.subs(eps,1)
def secular_DF_full(e,e1,w,w1,order):
    """ 
    Same as 'secular_DF' but uses the full expression for the Hansen coefficients that enter
    the disturbing function. While these expressions are typically more complicated when written
    in terms of eccentricity, they may be simpler when written in terms of canoncial momenta.
    """
    s = 0
    dw = w1 - w
    maxM = int( np.floor( order / 2.) )
    for i in range(maxM+1):
        s = s + secular_DF_harmonic_term(e,e1,dw,i,order) 
    return s

class laplace_B(Function):
    nargs=4
    @classmethod
    def eval(cls, s,j,n,alpha):
    # laplace coeffcients don't evaluate well with the C code
    # for large n, therefore I've substituted the exact expression
    # in terms of the elliptic K function when j=0
        if j is S.Zero:
            x = S('x')
            exprn = 4 / np.pi * diff(elliptic_k(x*x),x,n)
            return exprn.subs(x,alpha)
        elif j is S.One:
            x = S('x')
            exprn0 =  4 / np.pi *  (elliptic_k(x*x) - elliptic_e(x*x) ) / x 
            exprn = diff(exprn0, x ,n)
            return exprn.subs(x,alpha)
        else:
            return laplace_coefficient(s,j,n,alpha)/ alpha**n



from celmech import disturbing_function as df
import numpy as np

# taken from https://en.wikipedia.org/wiki/Farey_sequence
def farey_sequence(n):
    """Return the nth Farey sequence as order pairs of the form (N,D) where `N' is the numerator and `D' is the denominator."""
    a, b, c, d = 0, 1, 1, n
    sequence=[(a,b)]
    while (c <= n):
        k = int((n + b) / d)
        a, b, c, d = c, d, (k*c-a), (k*d-b)
        sequence.append( (a,b) )
    return sequence
def resonant_period_ratios(min_per_ratio,max_per_ratio,order):
    """Return the period ratios of all resonances up to order 'order' between 'min_per_ratio' and 'max_per_ratio' """
    if min_per_ratio < 0.:
        raise AttributeError("min_per_ratio of {0} passed to resonant_period_ratios can't be < 0".format(min_per_ratio))
    if max_per_ratio >= 1.:
        raise AttributeError("max_per_ratio of {0} passed to resonant_period_ratios can't be >= 1".format(max_per_ratio))
    minJ = int(np.floor(1. /(1. - min_per_ratio)))
    maxJ = int(np.ceil(1. /(1. - max_per_ratio)))
    res_ratios=[(minJ-1,minJ)]
    for j in range(minJ,maxJ):
        res_ratios = res_ratios + [ ( x[1] * j - x[1] + x[0] , x[1] * j + x[0]) for x in farey_sequence(order)[1:] ]
    res_ratios = np.array(res_ratios)
    msk = np.array( list(map( lambda x: min_per_ratio < x[0]/float(x[1]) < max_per_ratio , res_ratios )) )
    return res_ratios[msk]
def resonance_jk_list(min_per_ratio,max_per_ratio,order):
    """Return the 'j' and 'k' of resonances up to order 'order' between 'min_per_ratio' and 'max_per_ratio' """
    minJ = int(np.floor(1. /(1. - min_per_ratio)))
    maxJ = int(np.ceil(1. /(1. - max_per_ratio)))
    res_jk=[(minJ,1)]
    for j in range(minJ,maxJ):
        res_jk = res_jk + [ ( x[1] * j + x[0] , x[1]) for x in farey_sequence(order)[1:] ]
    res_jk =np.array(res_jk)
    msk = np.array( list(map( lambda x: min_per_ratio < float(x[0]-x[1])/float(x[0]) < max_per_ratio , res_jk )) )
    return res_jk[msk]
def resonance_pratio_span(mu1,mu2,Z0,res_j,res_k):
    """Compute the span of a resonance in terms of period ratio. 
    Inputs:
        mu1,mu2        --    Planet-star mass raitos of inner and outer planet
        Z0            --    Combined eccentricity of planet pair
        res_j,res_k    --    Specify the resonance as the res_j : (res_j-res_k) MMR
    Returns:
     the minimum and maximum period ratio spanned by resonance."""
    res_pratio = float(res_j - res_k) /float(res_j)
    alpha = res_pratio**(2./3.)
    beta = mu2 / mu1 / np.sqrt(alpha)
    f,g = df.get_fg_coeffs(res_j,res_k)

    # Pendulum approximation: 
    #        H(P,Q) =     (1/2)A P^2 + B cos(Q)
    A = 1.5 * (res_j*res_j + beta * res_j * (res_j-res_k))
    B = 2 * mu1 * np.sqrt(f*f + g*g)**res_k * Z0**res_k

    DeltaP = np.sqrt(4 * B / A )
    return (\
    res_pratio * (1. - 1.5 * res_j * DeltaP ) / (1 + 1.5 * beta * (res_j - res_k) * DeltaP ) , \
    res_pratio * (1. + 1.5 * res_j * DeltaP ) / (1 - 1.5 * beta * (res_j - res_k) * DeltaP )\
    )

def pendulum_approx_coeffs(mu1,mu2,res_j,res_k):
    res_pratio = float(res_j - res_k) /float(res_j)
    alpha = res_pratio**(2./3.)
    beta = mu2 / mu1 / np.sqrt(alpha)
    f,g = df.get_fg_coeffs(res_j,res_k)
    # Pendulum approximation: 
    #        H(P,Q) =     (1/2)A P^2 + B * Z0^k cos(Q)
    A = 1.5 * (res_j*res_j + beta * res_j * (res_j-res_k))
    B = 2 * mu1 * np.sqrt(f*f + g*g)**res_k
    return A,B,beta
    
def two_resonance_intersection(mu1,mu2,res_L,res_R):
    """Compute the intersection of adjacent resonances"""
    jL,kL = res_L
    jR,kR = res_R
    
    if (jL-kL)/float(jL) > (jR-kR)/float(jR):
    #swap left and right resonances
        jR,jL = jL,jR
        kR,kL = kL,kR
    max_k = np.max((kL,kR))
    #
    per_ratio_L = (jL-kL)/float(jL)
    per_ratio_R = (jR-kR)/float(jR)
    #
    aL,bL,betaL = pendulum_approx_coeffs(mu1,mu2,jL,kL)
    aR,bR,betaR = pendulum_approx_coeffs(mu1,mu2,jR,kR)
    #
    coeffL = per_ratio_L * 1.5 * (jL + betaL *(jL - kL)) * np.sqrt(4*bL/aL)
    coeffR = per_ratio_R * 1.5 * (jR + betaR *(jR - kR)) * np.sqrt(4*bR/aR)
    coeff0 = per_ratio_L - per_ratio_R
    #
    coeffs = np.zeros(max_k+1)
    coeffs[0] = coeff0
    coeffs[kL] = coeffL
    coeffs[kR] = coeffR
    roots = np.roots(coeffs[::-1])
    sqrt_e = np.min( roots[np.logical_and(np.imag(roots)==0,np.real(roots)>0)] ) 
    sqrt_e = np.real(sqrt_e)
    
    return (per_ratio_L + coeffL * sqrt_e**kL  , sqrt_e * sqrt_e)

def resonance_intersections_list(mu1,mu2,min_p,max_p,max_order):
    """
        Compute the list or resonance intersections for all resoances between 'min_per_ratio'
        and 'max_per_ratio' up to the specified 'order'. A list of ordered pairs containing 
        the period ratios and combined eccentricity values of the resonances intersections is
        returned. 
    """
    res_jk=resonance_jk_list(min_p,max_p,max_order)
    intersections=[two_resonance_intersection(mu1,mu2,r1,r2) for r1,r2 in zip(res_jk[:-1],res_jk[1:])]
    return np.array(intersections)
    



import numpy as np
# from celmech import Hamiltonian,sim_to_poincare_vars
import rebound


    
class multiplanetPoincareSystem(rebound.Simulation):
	
	
	def add(self, *args, **kwargs):

		super(multiplanetPoincareSystem, self).add(*args, **kwargs)
		self.sim_to_myvars()

	def sim_to_myvars(self):

		ps = self.particles
		Nps = len(ps)
		Mjac = np.zeros(Nps)
		mujac = np.zeros(Nps)
	
		Mstar = ps[0].m
		Mint = Mstar

		for p in ps[1:]:
			mi = p.m
			p.mjac = mi * Mint / (mi + Mint)
			p.Mjac =  Mstar * ( mi / p.mjac )
			p.mujac=sim.G**2 * p.Mjac**2 * p.mjac**3
			Mint += mi
			p.Lambda = p.mjac*np.sqrt(sim.G*p.Mjac*p.a)
			p.Gamma  = p.Lambda*(1.-np.sqrt(1.-p.e**2))
			p.lam  = p.l
			p.gamma  = -p.pomega

	def add_resonance(planet1,planet2,j,k,l):
		pass
	def add_all_resonances(planet1,planet2,j,k):
		pass
	def integrate(self,time):
		Nvariables = do_integration()
		self.assign_variables(Nvariables)
if __name__=="__main__":
	
	sim = multiplanetPoincareSystem()
	sim.add(m=1.)
	sim.add(m=1.e-5, a=3)
	sim.add(m=1.e-4, a=4)

	ps = sim.particles
	print( ps[1].m)
	print( ps[1].a)
	print( ps[1].Gamma)

# class 2bodysingleresonance(rebound.Simulation):
# 
# 	def add(self, *args, **kwargs):
# 		super(multiplanetSystem, self).add(args, kwargs)
# 		Nvariables = sim_to_myvars()
# 		self.assign_variables(Nvariables)
# 		self.particles[-1].Lambda = ...
# 	
# 	def add_resonance(planet1,planet2,j,k,l):
# 
# 	def integrate(self,time):
# 		
# 		Nvariables = do_integration()
# 
# 		self.assign_variables(Nvariables)



import numpy as np
from sympy import symbols, S, cos
from celmech.hamiltonian import Hamiltonian
from celmech.poincare import PoincareParticle, Poincare
from celmech.disturbing_function import get_fg_coeffs
from celmech.transformations import ActionAngleToXY, XYToActionAngle, pol_to_cart, cart_to_pol
import rebound
import warnings

# will give you the phiprime that yields an equilibrium Xstar, always in the range of phiprime where there exists a separatrix
def get_Phiprime(k, Xstarres):
    Phiprime = np.nan
    if Xstarres >= 0:
        warnings.warn("Xstarres passed to get_Phiprime must be 0. Returning nan")
    if k == 1:
        Phiprime = (4.*Xstarres**3 + 1.)/(3.*Xstarres)
    elif k == 2:
        Phiprime = (4.*Xstarres**2 - 2.)/3.
    elif k == 3:
        Phiprime = (4.*Xstarres**2 + 3.*Xstarres)/3.
    else:
        raise AttributeError('Order {0} not supported'.format(k))
    return Phiprime

def get_Xstarres(k, Phiprime): 
    Xstarres = np.nan
    if k == 1:
        if Phiprime >= 1:
            Xstarres = np.sqrt(Phiprime)*np.cos(1./3.*np.arccos(-Phiprime**(-1.5))+2.*np.pi/3.)
        elif Phiprime > 0:
            Xstarres = np.sqrt(Phiprime)*np.cosh(1./3.*np.arccosh(Phiprime**(-1.5)))
        elif Phiprime < 0:
            Xstarres = np.sqrt(-Phiprime)*np.sinh(1./3.*np.arcsinh(-abs(Phiprime)**(-1.5)))
    elif k == 2:
        if Phiprime < -2./3.: # wide of bifurcation. Origin is the stable fixed point
            Xstarres = 0.
        else:
            Xstarres = -0.5*np.sqrt(3.*Phiprime+2.)
    elif k == 3: 
        if Phiprime < -9./48:
            warnings.warn("Resonant fixed point for k=3 does not exist for Phiprime < -9./48.")
        else:
            Xstarres = (-3.-np.sqrt(9.+48.*Phiprime))/8.
    else:
        raise AttributeError('Order {0} not supported'.format(k))
    return Xstarres

def get_Xstarunstable(k, Phiprime):
    Xstarunstable = np.nan
    if k == 1:
        if Phiprime < 1.:
            warnings.warn("k=1 resonance has no unstable fixed point for Phiprime < 1")
        else:
            Xstarunstable = np.sqrt(Phiprime)*np.cos(1./3.*np.arccos(-Phiprime**(-1.5)))
    if k == 2:
        if Phiprime < -2./3.:
            warnings.warn("k=2 resonance has no unstable fixed point for Phiprime < -2/3")
        elif Phiprime < 2./3.:
            Xstarunstable = 0.
        else:
            Xstarunstable = 0.5*np.sqrt(3.*Phiprime-2.)
    if k == 3:
        if Phiprime < -9./48.:
            warnings.warn("k=3 resonance has no unstable fixed point for Phiprime < -9/48")
        else:
            Xstarunstable = (-3.+np.sqrt(9.+48.*Phiprime))/8.

    return Xstarunstable        

def get_Xstarnonres(k, Phiprime):
    Xstarnonres = np.nan
    if k == 1:
        if Phiprime < 1.:
            warnings.warn("k=1 resonance does not have a non-resonant fixed point for Phiprime < 1")
        else:
            Xstarnonres = np.sqrt(Phiprime)*np.cos(1./3.*np.arccos(-Phiprime**(-1.5)) + 4.*np.pi/3.)
    if k == 2:
        if Phiprime < 2./3.:
            warnings.warn("k=2 resonance has no non-resonant fixed point for Phiprime < 2/3")
        else:
            Xstarnonres = 0.
    if k == 3:
        Xstarnonres = 0.

    return Xstarnonres

def get_Hsep(k, Phiprime):
    Xu = get_Xstarunstable(k, Phiprime)
    Hsep = Xu**4 - 3.*Phiprime/2.*Xu**2 + np.abs(Xu)**(k-1)*Xu
    if np.isnan(Hsep):
        warnings.warn("There is no separatrix for the passed value of Phiprime")
    return Hsep

def get_Xsep(k, Phiprime):
    Xinner, Xouter = np.nan, np.nan
    if k==1:
        if Phiprime < 1.:
            warnings.warn("k=1 resonance has no separatrix for Phiprime < 1")
        else:
            Xu = get_Xstarunstable(k, Phiprime)
            disc = np.sqrt(1.5*Phiprime - 2*Xu**2)
            Xouter = -Xu - disc
            Xinner = -Xu + disc
    if k==2:
        if Phiprime < -2./3.:
            warnings.warn("k=2 resonance has no separatrix for Phiprime < -2/3")
        elif Phiprime < 2./3.:
            Xouter = -np.sqrt(3*Phiprime/2.+1.)
            Xinner = 0.
        else:
            Hsep = get_Hsep(k, Phiprime)
            b = (1.+1.5*Phiprime)
            disc = np.sqrt(b**2+4.*Hsep)
            Xouter = -np.sqrt((b+disc)/2.) # b >= 0 if Phiprime >= -2/3, so b + disc adds to give biggest absolute value
            Xinner = -np.sqrt((b-disc)/2.) # b >= 0 if Phiprime >= -2/3, so b - disc subtracts to give smaller absolute value
    if k==3:
        if Phiprime < -9./48.:
            warnings.warn("k=3 resonance has no separatrix for Phiprime < -9/48")
        else:
            Xu = get_Xstarunstable(k, Phiprime)
            disc = 1.+2*Xu
            Xouter = -(disc + np.sqrt(disc))/2.
            Xinner = -(disc - np.sqrt(disc))/2.
            #Xouter = -0.5*(1. + 2.*Xu + disc)
            #Xinner = -0.5*(1. + 2.*Xu - disc)
            
    return Xinner, Xouter

def get_num_fixed_points(k, Phiprime):
    num = 0
    if k==1:
        if Phiprime <= 1.:
            num = 1
        elif Phiprime > 1:
            num = 3
    if k==2:
        if Phiprime <= -2./3.:
            num = 1
        elif Phiprime <= 2./3.:
            num = 2
        elif Phiprime > 2/3.:
            num = 3
    if k==3:
        if Phiprime <= -9./48.:
            num = 1
        elif Phiprime > -9./48.:
            num = 3
    return num
            
class Andoyer(object):
    def __init__(self, j, k, X, Y, a10=1., G=1., m1=1.e-5, M1=1., m2=1.e-5, M2=1., B=1.5, Zcom=0., phiZcom=0., dKprime=0., theta=0., theta1=0.):
        self.calc_params(j, k, a10, G, m1, M1, m2, M2)
        self.X = X
        self.Y = Y
        self.Zcom = Zcom
        self.phiZcom = phiZcom
        self.B = B
        self.dKprime = dKprime
        self.theta = theta
        self.theta1 = theta1
    
    @property
    def Phi(self):
        Phi, phi = XYToActionAngle(self.X, self.Y)
        return Phi
    
    @property
    def phi(self):
        Phi, phi = XYToActionAngle(self.X, self.Y)
        return np.mod(phi, 2.*np.pi)

    @property
    def Phiprime(self):
        return 8.*self.B/3.
    @Phiprime.setter
    def Phiprime(self, value):
        self.B = 3.*value/8.

    @property
    def Zstar(self):
        Xstar = get_Xstarres(self.params['k'], self.Phiprime)
        Phistar = Xstar**2/2. 
        return self.Phi_to_Z(Phistar)
    
    @property
    def Zstar_unstable(self):
        Xstar_unstable = get_Xstarunstable(self.params['k'], self.Phiprime)
        Phistar_unstable = Xstar_unstable**2/2. 
        return self.Phi_to_Z(Phistar_unstable)
        
    @property
    def Zstar_nonres(self):
        Xstar_nonres = get_Xstarnonres(self.params['k'], self.Phiprime)
        Phistar_nonres = Xstar_nonres**2/2. 
        return self.Phi_to_Z(Phistar_nonres)

    @property
    def Zsep_outer(self):
        Xinner, Xouter = get_Xsep(self.params['k'], self.Phiprime)
        Phisep_outer = Xouter**2/2.
        return self.Phi_to_Z(Phisep_outer)
    
    @property
    def Zsep_inner(self):
        Xinner, Xouter = get_Xsep(self.params['k'], self.Phiprime)
        Phisep_inner = Xinner**2/2.
        return self.Phi_to_Z(Phisep_inner)
    
    @property
    def dP(self):
        return (self.Phi-self.B)*self.params['Phi0']

    @property
    def lambda1(self):
        p = self.params
        return np.mod((p['j']*p['K0']*self.theta1 - p['m2']*p['sLambda20']*self.theta)/p['K0'],2*np.pi)
    
    @property
    def lambda2(self):
        p = self.params
        return np.mod(((p['j']-p['k'])*p['K0']*self.theta1 + p['m1']*p['sLambda10']*self.theta)/p['K0'],2*np.pi)
        
    @property
    def psi1(self):
        return np.mod((self.phi - self.theta)/self.params['k'] ,2*np.pi)
    
    @property
    def Psi1(self):
        return self.Phi*self.params['k']*self.params['Phi0']
   
    @property
    def Z(self):
        return np.sqrt(2.*self.Psi1)/np.sqrt(self.params['Zfac'])
    
    @property
    def phiZ(self):
        return -self.psi1

    @property
    def Psi2(self):
        p = self.params
        return p['Zfac']*(p['C']*self.Zcom)**2/2.*p['K0']/p['eta']
    @property
    def psi2(self):
        return -(self.phiZcom + np.pi)
    
    @property
    def dL1hat(self):
        p = self.params
        return self.dKprime - (p['j']-p['k'])**2/3./p['j']*p['m2']*p['sLambda20']/p['K0']*self.dP

    @property
    def dL2hat(self):
        p = self.params
        return self.dKprime + (p['j']-p['k'])/3.*p['m1']*p['sLambda10']/p['K0']*self.dP
    
    @property
    def dK(self):
        return self.dKprime*self.params['K0']/self.params['eta']

    @property
    def wlib(self):
        return 2*(2*self.B)**(self.params['k']/4.)/self.params['tau']
    
    @property
    def tlib(self):
        return 2*np.pi/self.wlib

    def calc_params(self, j, k, a10, G, m1, M1, m2, M2):
        self.params = {'j':j, 'k':k, 'a10':a10, 'G':G, 'm1':m1, 'M1':M1, 'm2':m2, 'M2':M2}
        p = self.params
        p['alpha'] = (M1/M2)**(1./3.)*(float(j-k)/j)**(2./3.) 
        p['a20'] = a10/p['alpha']
        p['sLambda10'] = np.sqrt(G*M1*a10)
        p['sLambda20'] = np.sqrt(G*M2*a10/p['alpha'])
        p['K0'] = (j-k)*m2*p['sLambda20'] + j*m1*p['sLambda10']
        p['eta'] = float(j-k)/(3*j)*m1*m2*p['sLambda10']*p['sLambda20']/p['K0']
        
        f,g = get_fg_coeffs(j,k)
        p['f'], p['g'] = f,g
        norm = np.sqrt((j-k)/3./j)*np.sqrt((f**2*m2*p['sLambda20'] + g**2*m1*p['sLambda10'])/p['K0']) # np.sqrt(ff**2 + gg**2)
        p['Zfac'] = (f**2 + g**2)/norm**2
        p['C'] = -1./j*np.sqrt(g**2/(f**2 + g**2))*np.sqrt(float(j-k)/3./j)

        p['n0'] = 1./j*G**2*M1**2/p['sLambda10']**3
        p['a'] = -0.5*p['n0']*(p['j']-p['k'])**2
        p['c'] = -(j-k)*p['n0']*((j-k)/3./j)**((k-2.)/2.)*p['K0']/M2/np.sqrt(G*M1*a10)*((m2*p['sLambda20']*f**2 + m1*p['sLambda10']*g**2)/p['K0'])**(k/2.)
        p['Phi0'] = (4.*k**(k/2.)*p['c']/p['a'])**(2./(4.-k))
        p['tau'] = 4./(p['Phi0']*abs(p['a']))
    
    @classmethod
    def from_elements(cls, j, k, Zstar, libfac, a10=1., a1=None, G=1., m1=1.e-5, M1=1., m2=1.e-5, M2=1., Zcom=0., phiZcom=0., theta=0, theta1=0.):
        andvars = cls(j, k, 0., 0., a10=a10, G=G, m1=m1, M1=M1, m2=m2, M2=M2, Zcom=Zcom, phiZcom=phiZcom, dKprime=0., theta=theta, theta1=theta1)
        p = andvars.params
        Psi1star = 0.5*Zstar**2*p['Zfac']
        Phistar = Psi1star/k/p['Phi0']
        Xstar = -np.sqrt(2.*Phistar)
        andvars.Phiprime = get_Phiprime(k, Xstar)
        Xinner, Xouter = get_Xsep(k, andvars.Phiprime)
        #print(Zstar, Xstar, andvars.Phiprime, Xinner, Xouter)
        if libfac > 0: # offset toward outer branch of separatrix
            andvars.X = Xstar - libfac*np.abs(Xstar-Xouter)
        else: # offset toward inner branch of separatrix
            andvars.X = Xstar - libfac*np.abs(Xstar-Xinner)

        if a1 is None:
            a1 = a10
        dL1 = m1*np.sqrt(G*M1)*(np.sqrt(a1)-np.sqrt(a10))
        andvars.dKprime = p['eta']/p['m1']/p['sLambda10']/p['eta']*(dL1 + (j-k)*p['eta']*andvars.dP)
        
        return andvars
    
    @classmethod
    def from_Z(cls, j, k, Z, phi, Zstar, a10=1., G=1., m1=1.e-5, M1=1., m2=1.e-5, M2=1., Zcom=0., phiZcom=0., dKprime=0., theta=0, theta1=0.):
        andvars = cls(j, k, 0., 0., a10=a10, G=G, m1=m1, M1=M1, m2=m2, M2=M2, Zcom=Zcom, phiZcom=phiZcom, dKprime=dKprime, theta=theta, theta1=theta1)
        Phistar = andvars.Z_to_Phi(Zstar)
        Xstar = -np.sqrt(2.*Phistar)
        andvars.Phiprime = get_Phiprime(k, Xstar)
        
        Phi = andvars.Z_to_Phi(Z)
        andvars.X = np.sqrt(2.*Phi)*np.cos(phi)
        andvars.Y = np.sqrt(2.*Phi)*np.sin(phi)

        return andvars

    @classmethod
    def from_dP(cls, j, k, dP, phi, Zstar, a10=1., G=1., m1=1.e-5, M1=1., m2=1.e-5, M2=1., Zcom=0., phiZcom=0., dKprime=0., theta=0, theta1=0.):
        andvars = cls(j, k, 0., 0., a10=a10, G=G, m1=m1, M1=M1, m2=m2, M2=M2, Zcom=Zcom, phiZcom=phiZcom, dKprime=dKprime, theta=theta, theta1=theta1)
        Phistar = andvars.Z_to_Phi(Zstar)
        Xstar = -np.sqrt(2.*Phistar)
        andvars.Phiprime = get_Phiprime(k, Xstar)

        Phi = andvars.dP_to_Phi(dP, andvars.Phiprime)
        andvars.X = np.sqrt(2.*Phi)*np.cos(phi)
        andvars.Y = np.sqrt(2.*Phi)*np.sin(phi)

        return andvars
    
    @classmethod
    def from_Poincare(cls, pvars, j, k, a10, i1=1, i2=2):
        p1 = pvars.particles[i1]
        p2 = pvars.particles[i2]
        andvars = cls(j, k, 0., 0., a10=a10, G=p1.G, m1=p1.m, M1=p1.M, m2=p2.m, M2=p2.M)
        
        p = andvars.params
        dL1hat = (p1.sLambda-p['sLambda10'])/p['sLambda10']
        dL2hat = (p2.sLambda-p['sLambda20'])/p['sLambda20']

        dP = 3.*j/(j-k)*(dL2hat - dL1hat)
        andvars.dKprime = ((j-k)*p['m2']*p['sLambda20']*dL2hat + j*p['m1']*p['sLambda10']*dL1hat)/p['K0']
       
        andvars.theta = j*p2.l - (j-k)*p1.l
        andvars.theta1 = (p['m1']*p['sLambda10']*p1.l + p['m2']*p['sLambda20']*p2.l)/p['K0']
      
        Z, phiZ, andvars.Zcom, andvars.phiZcom = andvars.sGammas_to_Zs(p1.sGamma, p1.gamma, p2.sGamma, p2.gamma)
        Psi1 = p['Zfac']*Z**2/2.
        psi1 = -phiZ

        andvars.B = (Psi1/k - dP)/p['Phi0']
        Phi = Psi1/k/p['Phi0']
        phi = andvars.theta + k*psi1

        andvars.X = np.sqrt(2*Phi)*np.cos(phi)
        andvars.Y = np.sqrt(2*Phi)*np.sin(phi)
        
        return andvars

    def to_Poincare(self):
        p = self.params
        
        sLambda1 = p['sLambda10']*(1.+self.dL1hat)
        sLambda2 = p['sLambda20']*(1.+self.dL2hat)

        sGamma1, gamma1, sGamma2, gamma2 = self.Zs_to_sGammas(self.Z, self.phiZ, self.Zcom, self.phiZcom)
        
        pvars = Poincare(p['G'])
        pvars.add(m=p['m1'], sLambda=sLambda1, l=self.lambda1, sGamma=sGamma1, gamma=gamma1, M=p['M1'])
        pvars.add(m=p['m2'], sLambda=sLambda2, l=self.lambda2, sGamma=sGamma2, gamma=gamma2, M=p['M2'])

        return pvars

    @classmethod
    def from_Simulation(cls, sim, j, k, a10=None, i1=1, i2=2, average=True):
        if a10 is None:
            a10 = sim.particles[i1].a
        pvars = Poincare.from_Simulation(sim, average)
        #if average is True:
        #    pvars.average_resonant_terms(i1=i1, i2=i2, exclude=[[j,k]])
        
        return Andoyer.from_Poincare(pvars, j, k, a10, i1, i2)

    def to_Simulation(self, masses=None, average=True):
        ''' By default assumes 2 planet system with self consistent jacobi masses
        (4 masses m1, M1, m2, M2 for 3 masses of star+2 planets may not be consistent).
        If 2 planets are part of larger system, need to pass physical masses=[M, m1, m2]
        '''
        pvars = self.to_Poincare()
        #if average is True:
        #    pvars.average_resonant_terms(exclude=[[self.params['j'], self.params['k']]], inverse=True)
        return pvars.to_Simulation(masses, average)

    def Z_to_Phi(self, Z):
        p = self.params
        Psi1 = 0.5*Z**2*p['Zfac']
        Phi = Psi1/p['k']/p['Phi0']
        return Phi
    
    def Phi_to_Z(self, Phi):
        p = self.params
        Psi1 = Phi*p['Phi0']*p['k']
        Z = np.sqrt(2*Psi1)/np.sqrt(p['Zfac'])
        return Z 
    
    def dP_to_Phi(self, dP, Phiprime):
        p = self.params
        B = 3.*Phiprime/8.
        Phi = (dP + B)/p['Phi0']
        return Phi
    
    def Phi_to_dP(self, Phi, Phiprime):
        p = self.params
        B = 3.*Phiprime/8.
        dP = Phi*p['Phi0'] - B
        return dP 

    def Zs_to_sGammas(self, Z, phiZ, Zcom, phiZcom):
        ZX, ZY = pol_to_cart(Z, phiZ)
        ZcomX, ZcomY = pol_to_cart(Zcom, phiZcom)
        p = self.params
        ratio = np.sqrt(p['M2']*p['a20']/p['M1']/p['a10'])
        m1, m2 = p['m1'], p['m2']
        f, g = p['f'], p['g']
        beta1 = abs(f/g)*ratio
        beta2 = float(p['j']-p['k'])/p['j']*ratio
        D = m1*g - f*beta1*m2
        N = np.sqrt(f**2 + g**2)
        sX1 = -np.sqrt(p['sLambda10'])/D*(beta1*m2*N*ZX - g*(m1+beta2*m2)*ZcomX)
        sY1 = -np.sqrt(p['sLambda10'])/D*(-beta1*m2*N*ZY + g*(m1+beta2*m2)*ZcomY)
        sX2 = np.sqrt(p['sLambda20'])/D*(m1*N*ZX - f*(m1+beta2*m2)*ZcomX)
        sY2 = np.sqrt(p['sLambda20'])/D*(-m1*N*ZY + f*(m1+beta2*m2)*ZcomY)
        sGamma1, gamma1 = XYToActionAngle(sX1, sY1)
        sGamma2, gamma2 = XYToActionAngle(sX2, sY2)
        return sGamma1, gamma1, sGamma2, gamma2

    def sGammas_to_Zs(self, sGamma1, gamma1, sGamma2, gamma2):
        sX1, sY1 = ActionAngleToXY(sGamma1, gamma1)
        sX2, sY2 = ActionAngleToXY(sGamma2, gamma2)
        p = self.params
        ratio = np.sqrt(p['M2']*p['a20']/p['M1']/p['a10'])
        m1, m2 = p['m1'], p['m2']
        f, g = p['f'], p['g']
        beta1 = abs(f/g)*ratio
        beta2 = float(p['j']-p['k'])/p['j']*ratio
        N = np.sqrt(f**2 + g**2)
        ZX = 1/N*(f/np.sqrt(p['sLambda10'])*sX1 + g/np.sqrt(p['sLambda20'])*sX2)
        ZY = 1/N*(-f/np.sqrt(p['sLambda10'])*sY1 - g/np.sqrt(p['sLambda20'])*sY2)
        ZcomX = 1/(m1 + beta2*m2)*(m1/np.sqrt(p['sLambda10'])*sX1 + beta1*m2/np.sqrt(p['sLambda20'])*sX2)
        ZcomY = 1/(m1 + beta2*m2)*(-m1/np.sqrt(p['sLambda10'])*sY1 - beta1*m2/np.sqrt(p['sLambda20'])*sY2)
        Z, phiZ = cart_to_pol(ZX, ZY)
        Zcom, phiZcom = cart_to_pol(ZcomX, ZcomY)
        return Z, phiZ, Zcom, phiZcom
    
class AndoyerHamiltonian(Hamiltonian):
    def __init__(self, andvars):
        X, Y, Phi, phi, B, theta, k = symbols('X, Y, Phi, phi, B, theta, k')
        pqpairs = [(X, Y), (B, theta)]
        p = andvars.params
        Hparams = {k:p['k']}

        H = -((X**2 + Y**2) - S(2)*B)**2 - (X**2 + Y**2)**((k-S(1))/S(2))*X # + n0*tau*dK*(1-1.5*eta*Phi0*dK/K0)
        
        super(AndoyerHamiltonian, self).__init__(H, pqpairs, Hparams, andvars)  
        self.Hpolar = -4*(Phi-B)**2 - (2*Phi)**(k/S(2))*cos(phi)

    def state_to_list(self, state):
        return [state.X, state.Y, state.B, -state.theta] # - because B is conjugate to -theta

    def update_state_from_list(self, state, y):
        p = state.params
        state.X = y[0]
        state.Y = y[1]
        state.B = y[2]
        state.theta = -y[3] # because B conjugate to -theta

        theta1dot = p['n0']*p['tau']*(1.-3.*p['Phi0']*state.dKprime)
        state.theta1 = theta1dot*self.integrator.t



from sympy import S, diff, lambdify, symbols 
from scipy.integrate import ode

class Hamiltonian(object):
    def __init__(self, H, pqpairs, Hparams, initial_state):
        """
        H = Hamiltonian made up only of sympy symbols in pqpairs and keys in Hparams
        pqpairs = list of momentum, position pairs [(P1, Q1), (P2, Q2)], where each element is a sympy symbol
        Hparams = dictionary from sympy symbols for the constant parameters in H to their value
        initial_conditions = arbitrary object for holding the dynamical state

        The variables are stored in the self.state object, while ODE uses a list of dynamical variables y. 
        In addition to the above, one needs to write 2 methods to map between the two objects:
        def state_to_list(self, state): 
            returns a list of values from state in the same order as pqpairs e.g. [P1,Q1,P2,Q2]
        def update_state_from_list(self, state, y):
            updates state object from a list of values y for the variables in the same order as pqpairs
        """
        self.state = initial_state
        self.Hparams = Hparams
        self.H = H
        self.pqpairs = pqpairs
        self.varsymbols = [var for pqpair in self.pqpairs for var in pqpair]
        self._update()

    def integrate(self, time):
        try:
            time /= self.state.params['tau'] # scale time if defined
        except:
            pass
        if not hasattr(self, 'Nderivs'):
            self._update()
        if time > self.integrator.t:
            try:
                self.integrator.integrate(time)
            except:
                raise AttributeError("Need to initialize Hamiltonian")
        self.update_state_from_list(self.state, self.integrator.y)

    def _update(self):
        self.NH = self.H # reset to Hamiltonian with all parameters unset
        # 
        # Update raw numerical constants first then update functions
        # Less hacky way to do this?
        function_keyval_pairs = []
        for key, val in self.Hparams.items(): 
            if isinstance(val,float):
                self.NH = self.NH.subs(key, val)
            else:
                function_keyval_pairs.append((key,val)) 
        for keyval in function_keyval_pairs:
            self.NH = self.NH.subs(keyval[0],keyval[1])
        
        self.derivs = {}
        self.Nderivs = []
        for pqpair in self.pqpairs:
            p = pqpair[0]
            q = pqpair[1]
            self.derivs[p] = -diff(self.H, q)
            self.derivs[q] = diff(self.H, p)
            self.Nderivs.append(lambdify(self.varsymbols, -diff(self.NH, q), 'numpy'))
            self.Nderivs.append(lambdify(self.varsymbols, diff(self.NH, p), 'numpy'))

        def diffeq(t, y):
            dydt = [deriv(*y) for deriv in self.Nderivs]
            #print(t, y, dydt)
            return dydt
        self.integrator = ode(diffeq).set_integrator('lsoda')
        self.integrator.set_initial_value(self.state_to_list(self.state))



from celmech import disturbing_function
import numpy as np
from collections import OrderedDict
from sympy import S
from celmech.disturbing_function import laplace_coefficient 
from celmech.disturbing_function import get_fg_coeffs
from itertools import combinations
import rebound

def masses_to_jacobi(masses):
    N = len(masses)
    mjac, Mjac = np.zeros(N), np.zeros(N)
    interior_mass = masses[0]
    for i in range(1,N):
        mjac[i] = masses[i]*interior_mass/(masses[i]+interior_mass) # reduced mass with interior mass
        Mjac[i] = masses[0]*(masses[i]+interior_mass)/interior_mass # mjac[i]*Mjac[i] always = m[i]*m[0]
        interior_mass += masses[i]
    mjac[0] = interior_mass
    return list(mjac), list(Mjac)

def masses_from_jacobi(mjac, Mjac):
    plus = Mjac[1] # m0+m1
    product = mjac[1]*Mjac[1] # m0m1
    minus = np.sqrt(plus**2 - 4*product) # m0-m1
    m0 = (plus + minus)/2.
    masses = [m0] + [Mjac[i]*mjac[i]/m0 for i in range(1,len(mjac))] # mjac[i]*Mjac[i] = m0mi
    return masses

def ActionAngleToXY(Action,angle):
        return np.sqrt(2*Action)*np.cos(angle),np.sqrt(2*Action)*np.sin(angle)

def XYToActionAngle(X,Y):
        return 0.5 * (X*X+Y*Y), np.arctan2(Y,X)

def pol_to_cart(R, phi):
    X = R*np.cos(phi)
    Y = R*np.sin(phi)
    return X, Y

def cart_to_pol(X, Y):
    R = np.sqrt(X**2 + Y**2)
    phi = np.arctan2(Y,X)
    return R, phi

